#!/usr/bin/env python

from pyparsing import oneOf, Literal, Regex, Suppress, Optional, OneOrMore, Forward, SkipTo, ZeroOrMore, restOfLine, Empty,Group, Combine, ParseException
from GenericParser import *
import re



class PokerStarsNLHECashGameParser(GenericParser):

    def __init__(self, Dispatcher=None):
        super(PokerStarsNLHECashGameParser,  self).__init__()
        # each parser should register the game type it recognises with the Dispatcher
        # by calling ...xxx... with an identifying regex.

        # Identifies games of type:
        # PokerStars Game #17955774161:  Hold'em No Limit ($0.01/$0.02) - 2008/06/06 - 13:20:26 (ET)
        if Dispatcher is not None:
            Dispatcher.register(self,  "PokerStars Game #\d+:  Hold'em No Limit \(\$?\d+\.(\d\d)?\/\$?\d+\.(\d\d?)\) - \d\d\d\d\/\d\d\/\d\d - \d\d:\d\d:\d\d \(ET\)")


    # ParseActions
    # These translate the tokens from the PokerStars parser
    # to generic calls on the GenericParser parent class, which builds the Hand (or prints stuff, depending)

    def newHand(self,s,l,t):
        startTime = t['startTime']
        handNumber = t['handNumber']
        smallblind = t['smallblind']
        bigblind = t['bigblind']
        GenericParser.newHand(self, startTime=startTime, handNumber=handNumber,  smallblind=smallblind,  bigblind=bigblind)

    def addPlayer(self,s,l,t):
        (name, chips, seat) = (t['playerName'], t['playerChips'], int(t['playerSeat']))
        GenericParser.addPlayer(self, name=name, chips=chips, seat=seat)

    def datify(self,s,l,t):
        """extracts the individual time components from a Regex parse element of a date/time and returns a [QDateTime] in their place."""
        #t['startTime'] = QDateTime(QDate(int(t['yyyy']),int(t['mm']),int(t['dd'])), QTime(int(t['h']),int(t['m']),int(t['s'])))
        #t['startTime'] = "%4d/%2d/%2d %2d:%2d:%2d" % (t['yyyy'], t['mm'], t['dd'], t['h'], t['m'], t['s'])
        t['startTime'] = "%s/%s/%s %s:%s:%s" % (t['yyyy'], t['mm'], t['dd'], t['h'], t['m'], t['s'])
        
    def monify(self,s,l,t):
        """turns a string of money into a Decimal"""
        amount = re.sub("\$", '', t['amount']) # consider precompiling the re
        return (Decimal(amount))

    def setButton(self,s,l,t):
        buttonSeat = int(t['buttonseat'])
        GenericParser.setButton(self, buttonSeat)

    def setHoleCards(self,s,l,t):
        #SHOWS = Group(PLAYER + Suppress(": shows [") + Group(CARD + CARD) + Suppress("] (") + HAND + Suppress(")"))
        #Group(Suppress("Dealt to") + PLAYER + Suppress("[") + Group(CARD + CARD) + Suppress("]"))
        hole = t[0][1]
        name = t[0][0]
        GenericParser.setHoleCards(self, name, hole)

    def setFlop(self,s,l,t):
        flopCards = t[0]
        GenericParser.setFlop(self, flopCards)

    def setTurn(self,s,l,t):
        turnCard = t[0][0]
        GenericParser.setTurn(self, turnCard)

    def setRiver(self,s,l,t):
        riverCard = t[0][0]
        GenericParser.setRiver(self, riverCard)

    def addBlind(self,s,l,t):
        name = t[0][0]
        amount = t[0][1]
        GenericParser.addBlind(self, name, amount)

    def addCall(self,s,l,t):
        name = t[0][0]
        amount = t[0][1]
        GenericParser.addCall(self, name, amount)

    def addCheck(self,s,l,t):
        name = t[0][0]
        GenericParser.addCheck(self, name)

    def addBet(self,s,l,t):
        name = t[0][0]
        amount = t[0][1]
        GenericParser.addBet(self, name, amount)

    def addRaise(self,s,l,t):
        name = t[0][0]
        amount = t[0][2]
        GenericParser.addRaise(self, name, amount)

    def addFold(self,s,l,t):
        name = t[0][0]
        GenericParser.addFold(self, name)

    def uncalledBet(self,s,l,t):
        name = t[0][1]
        amount = t[0][0]
        GenericParser.addUncalled(self, name, amount)

    def collectedAmount(self,s,l,t):
        name = t[0][0]
        amount = t[0][1]
        GenericParser.collectedAmount(self, name, amount)

    def wonAmount(self,s,l,t):
        name =  t[0][0]
        amount =  t[0][1]
        GenericParser.collectedAmount(self, name, amount)

    def setRake(self,s,l,t):
        amount = t[0][0]
        GenericParser.setRake(self, amount)

    def endHand(self,s,l,t):
        GenericParser.endHand(self)

    def addMuckedCards(self,s,l,t):
        #    MUCKED = Group(self.PLAYER + Suppress(Optional(BLINDPOS) + "mucked [") + Group(CARD + CARD) + Suppress("]"))
        name = t[0][0]
        hole = t[0][1]
        GenericParser.addMuckedCards(self, name, hole)

    def parser(self):
        """PokerStars No Limit Hold 'Em cash game importer. Returns the grammar for a single PokerStars NLHE cash game."""

        PLAYER = self.PLAYER

        #MONEY = Optional("$") + Regex("(?P<amount>\d+(\.\d\d)?)")
        MONEY = Regex("(?P<amount>\$?\d+(\.\d\d)?)").setParseAction(self.monify)


        CARD = Combine(oneOf("A K Q J T 9 8 7 6 5 4 3 2") + oneOf("s h d c"))
        CARDNAME = Literal("Ace") | Literal("King") | Literal("Queen") | Literal("Jack") | Literal("Ten") | Literal("Nine") | Literal("Eight") | Literal("Seven") | Literal("Six") | Literal("Five") | Literal("Four") | Literal("Three") | Literal("Deuce")
        CARDSNAME = Literal("Aces") | Literal("Kings") | Literal("Queens") | Literal("Jacks") | Literal("Tens") | Literal("Nines") | Literal("Eights") | Literal("Sevens") | Literal("Sixes") | Literal("Fives") | Literal("Fours") | Literal("Threes") | Literal("Deuces")

        HIGHCARD = Literal("high card") + CARDNAME + Optional(Regex(".+kicker"))
        PAIR = Literal("a pair of") + CARDSNAME + Optional(Regex(".+kicker"))
        TWOPAIR = Literal("two pair,") + CARDSNAME + Literal("and") + CARDSNAME + Optional(Regex(".+kicker"))
        SET = Literal("three of a kind,") + CARDSNAME + Optional(Regex(".+kicker"))
        STRAIGHT = Literal("a straight,") + CARDNAME + Literal("to") + CARDNAME
        FLUSH = Literal("a flush,") + CARDNAME + Literal("high")
        QUAD = Literal("four of a kind,") + CARDSNAME + Optional(Regex(".+kicker"))
        FULLHOUSE = Literal("a full house,") + CARDSNAME + "full of" + CARDSNAME
        STRAIGHTFLUSH = Literal("a straight flush,") + CARDNAME + "to" + CARDNAME
        ROYALFLUSH = Literal("a Royal Flush")

        HAND = HIGHCARD | PAIR | TWOPAIR | SET | QUAD | STRAIGHT | FLUSH | FULLHOUSE | STRAIGHTFLUSH | ROYALFLUSH


        # miscellaneous actions that seems to occur almost anywhere.
        # since these may be performed by people not sitting at the table, we cannot be sure that they will match a PLAYER. We pretty much ignore them.
        LEAVES = Regex(".+ leaves the table")
        HASTIMEDOUT = Regex(".+ has timed out( while disconnected)?")
        ISSITTINGOUT = Regex(".+(: )?is sitting out")
        JOINS = Regex(".+ joins the table at seat #\d+")
        HASRETURNED = Regex(".+ has returned")
        ISDISCONNECTED = Regex(".+ is disconnected")
        ISCONNECTED = Regex(".+ is connected")
        SAID = Regex('.+ said, ".+"')
        WASREMOVED = Regex(".+ was removed from the table for failing to post")
        AFTERBUTTON = Regex(".+ will be allowed to play after the button")
        SITSOUT = Regex(".+: sits out")
        MISC = LEAVES | HASTIMEDOUT | ISSITTINGOUT | JOINS | HASRETURNED | SAID | ISDISCONNECTED | ISCONNECTED | WASREMOVED | SITSOUT | AFTERBUTTON


        # blinds.
        PAYSBLIND = Group(self.PLAYER + Suppress(Regex(": posts (small blind|big blind|small \& big blinds)")) + MONEY  + Optional("and is all-in"))
        PAYSBLIND.setParseAction(self.addBlind) 
        BLIND = PAYSBLIND | MISC
        
        # betting actions.
        FOLDS = Group(self.PLAYER + Suppress(": folds")).setParseAction(self.addFold)
        CALLS = Group(self.PLAYER + Suppress(": calls") + MONEY + Optional("and is all-in"))
        CALLS.setParseAction(self.addCall)
        RAISES = Group(self.PLAYER + Suppress(": raises") + MONEY + Suppress("to") + MONEY + Optional("and is all-in"))
        RAISES.setParseAction(self.addRaise)
        CHECKS = Group(self.PLAYER + Suppress(": checks"))
        CHECKS.setParseAction(self.addCheck)
        BETS = Group(self.PLAYER + Suppress(": bets") + MONEY + Optional("and is all-in"))
        BETS.setParseAction(self.addBet)
        UNCALLED = Group(Suppress("Uncalled bet (") + MONEY + Suppress(") returned to") + self.PLAYER)
        UNCALLED.setParseAction(self.uncalledBet)
        BETCOLLECTED = Group(self.PLAYER + Suppress("collected") + MONEY + Suppress("from" + restOfLine))
        #COLLECTED.setParseAction(self.collected)
        DOESNTSHOW = Group(self.PLAYER + Suppress(":") + "doesn't show hand")
        SHOWS = Group(PLAYER + Suppress(": shows [") + Group(CARD + CARD) + Suppress("] (") + HAND + Suppress(")"))
        SHOWS.setParseAction(self.setHoleCards)
        BET = FOLDS | CALLS | RAISES | CHECKS | BETS | MISC | UNCALLED | BETCOLLECTED | DOESNTSHOW | SHOWS


        # showdown actions.
        MUCKS = self.PLAYER + ": mucks hand"
        SHOWDOWNACTION = SHOWS | DOESNTSHOW | MUCKS | BETCOLLECTED | MISC

        #
        # SUMMARY
        #
        # The summary must be parsed as it is the only place where
        # mucked hands are sometimes reported.
        # Also we get all winnings from here, handily collated per seat
        MAINPOT = "Main pot" + MONEY + "."
        SIDEPOT = Regex("Side pot(-\d)?") + MONEY + "."
        RAKE = Group(MONEY)
        POT = "Total pot" + MONEY + Optional(MAINPOT + OneOrMore(SIDEPOT)) + "| Rake" + RAKE
        RAKE.setParseAction(self.setRake)
        BOARD = "Board [" + ZeroOrMore(CARD) + "]"
        BLINDPOS = Optional(Literal("(small blind)") | Literal("(big blind)") | Literal("(small & big blinds)")) + Optional("(button)")
        MUCKED = Group(self.PLAYER + Suppress(Optional(BLINDPOS) + "mucked [") + Group(CARD + CARD) + Suppress("]"))
        MUCKED.setParseAction(self.addMuckedCards)
        WON = Group(self.PLAYER  + Suppress(Optional(BLINDPOS) + "showed [" + Group(CARD + CARD) + "] and won (") + MONEY + Suppress(") with") + HAND)
        COLLECTED = Group(self.PLAYER + Suppress(Optional(BLINDPOS) + "collected (") + MONEY + Suppress(")"))
        WON.setParseAction(self.wonAmount)
        COLLECTED.setParseAction(self.collectedAmount)
        SUMMARYLINE = Regex("Seat \d+:") + (MUCKED | WON | COLLECTED | restOfLine) # note : quiety matches lots of summary lines, maybe warn about this. (eg if mucked but not showing cards -- which is something I'm interested in, as you can obtain range of losing hands)
        


        SEAT = Regex("Seat (?P<playerSeat>\d+): (?P<playerName>.+) \(") + MONEY.setResultsName('playerChips') + Suppress("in chips)")
        SEAT.setParseAction(self.addPlayer)


        # The main stages of a hand
        HOLECARDS = Group(Suppress("Dealt to") + PLAYER + Suppress("[") + Group(CARD + CARD) + Suppress("]"))
        HOLECARDS.setParseAction(self.setHoleCards)
        PREFLOP = "*** HOLE CARDS ***" + Optional(HOLECARDS) + ZeroOrMore(BET)
        
        FLOP = Suppress("*** FLOP *** [") + Group(CARD + CARD + CARD).setParseAction(self.setFlop) + Suppress("]") + Group(ZeroOrMore(BET))
        #FLOP.setParseAction(self.setFlop)
        TURN = Suppress("*** TURN *** [" + CARD + CARD + CARD + "] [") + Group(CARD).setParseAction(self.setTurn) + Suppress("]") + Group(ZeroOrMore(BET))
        SHOWDOWN = "*** SHOW DOWN ***" + OneOrMore(SHOWDOWNACTION)
        RIVER = Suppress("*** RIVER *** [" + CARD + CARD + CARD + CARD + "] [") + Group(CARD).setParseAction(self.setRiver) + Suppress("]") + Group(ZeroOrMore(BET))
        SUMMARY = "*** SUMMARY ***" + Optional(POT) + Optional(BOARD) + OneOrMore(SUMMARYLINE)
        SUMMARY.setParseAction(self.endHand)
        BETTING = OneOrMore(BLIND) + (Optional(PREFLOP + Optional(FLOP) + Optional(TURN) + Optional(RIVER) + Optional(SHOWDOWN)) | "Hand cancelled")

        STARTTIME = Regex("(?P<yyyy>\d\d\d\d)\/(?P<mm>\d\d)\/(?P<dd>\d\d) - (?P<h>\d\d):(?P<m>\d\d):(?P<s>\d\d) \(ET\)").setParseAction(self.datify)
        GTSNG = Regex("Tournament")
        GTHOLDEM = "Hold'em No Limit (" + MONEY.setResultsName('smallblind') + "/" + MONEY.setResultsName('bigblind') + Suppress(") -") + STARTTIME

        GAMETYPE = GTHOLDEM | GTSNG
        HEAD = Regex("PokerStars Game #(?P<handNumber>\d+):") + GAMETYPE
        HEAD.setParseAction(self.newHand)
        TABLE = Regex("Table '.+' \d+-max Seat #(?P<buttonseat>\d+) is the button").setParseAction(self.setButton)


        # an entire hand.
        GAME = HEAD + TABLE + OneOrMore(SEAT) + BETTING + SUMMARY


        # unused; taken care of by an external loop.
        #SESSION = OneOrMore(GAME)
        return GAME




if __name__ == '__main__':
    #main()
    parser = PokerStarsNLHECashGameParser()
    fname = "PokerStarsNLHECashGame.sample.txt"
    fh = None
    try:
        fh = open(fname)
        history = fh.read()
    finally:
        if fh is not None:
            fh.close()

    parser.parseSession(history)


