from decimal import *
import operator

class Session(object):
    def __init__(self):
        self.players = []   # list of players over the entire session
        self.hands = []     # list of Hands
        self.startTime = None
        self.endTime = None
        self.chips = {}     # dict from names to cumulative chip stacks
        self.numHands = {}
        self.blinds = {}

    def addPlayer(self, seat=None, name=None, chips=None):
        if seat and name and chips:
            # if the player is new to the session, add him to players and start to track his stack over the session.
            if name not in self.players:
                self.players.append(name)
                self.chips[name] = 0
                self.numHands[name] = 0
                self.blinds[name] = 0
            # add him to the current game,
            # perhaps sanity checking that he hasn't mysteriously changed seat, and flagging whether his stack changed unexpectedly indicating error or rebuy
            self.currHand.addPlayer(seat,name,chips)
            if len(self.hands) > 1:
                startchipsthishand = self.currHand.startChips[name]
                if name in self.hands[-2].endChips.keys():
                    endchipslasthand = self.hands[-2].endChips[name]
                    if not startchipsthishand == endchipslasthand :
                        print "unexplained stack change: last hand,", endchipslasthand, "this hand", startchipsthishand, "difference:", startchipsthishand - endchipslasthand, name
            self.numHands[name]+=1
            
            # seat number?


    def newHand(self,handNumber,startTime):
        self.hands.append(Hand(handNumber,startTime))
        self.currHand = self.hands[-1]

    def endHand(self):
        # at the end of each hand
        netChipChange = self.currHand.endHand()
        for name in netChipChange.keys():
            self.chips[name] += netChipChange[name]


    def addBlind(self, name, amount):
        self.blinds[name] += amount
        self.currHand.addBlind(name,amount)

    def endSession(self):
        self.startTime = self.hands[0].startTime
        self.endTime = self.hands[len(self.hands)-1].startTime

class Hand(object):

    (PREFLOP, FLOP, TURN, RIVER) = range(4)

    def __init__(self, handNumber, startTime):
        self.handNumber = handNumber # the site's handnumber
        self.startTime = startTime   # when the hand was dealt
        self.players = []   #  list of players' names dealt into the hand
        self.startChips = {} # chips per player at start, indexed by name
        self.endChips = {}   # chips per player at end, indexed by name
        self.holeCards = {}  # holecards if seen, indexed by name
        self.button = None # if it's still None later on, there's been a problem
        self.winners = {}
        self.netWinners = {}
        self.rake = 0
        # board cards
        self.flop = None
        self.turn = None
        self.river = None
        self.street = Hand.PREFLOP  # state variable, used during construction
        # bets is a list of dicts, one per street, from player names to lists of bets
        self.bets = [{},{},{},{}]


    def addPlayer(self, seat, name, chips):
        """seat, an int indicating the seat
        name, the player name
        chips, the chips the player has at the start of the hand"""
        self.players.append(name)
        self.startChips[name] = chips
        self.endChips[name] = chips
        self.winners[name] = 0
        for street in range(4):
            self.bets[street][name] = []

    def setButton(self,button):
        """ button, an int giving the seat number of the button"""
        self.button=button

    def setHoleCards(self,player,holeCards):
        if player in self.holeCards.keys():
            # we already saw them, perhaps because they were dealt to us, lets just check they haven't changed!
            if not self.holeCards[player] == holeCards:
                print("parse error: our holecards changed during a hand")
                print "original: ", self.holeCards[player]
                print "new     : ", holeCards
        else:
            self.holeCards[player] = holeCards

    def setFlop(self, flop):
        self.flop = frozenset(flop)
        self.street = Hand.FLOP

    def setRiver(self, river):
        self.river = river
        self.street = Hand.RIVER

    def setTurn(self, turn):
        self.turn = turn
        self.street = Hand.TURN

    def addBlind(self, name, amount):
        self.bets[self.street][name].append(amount)

    def addCall(self, name, amount):
        self.bets[self.street][name].append(amount)

    def addBet(self, name, amount):
        self.bets[self.street][name].append(amount)

    def addRaise(self, name, amount):
        # amount is the amount raised to, not the amount raised.
        committedthisstreet = reduce(operator.add, self.bets[self.street][name], 0)
        self.bets[self.street][name].append(amount - committedthisstreet)

    def addFold(self,name):
        self.bets[self.street][name].append(0)
        pass # hah

    def addUncalled(self, name, amount):
        self.bets[self.street][name].append(-amount)

    def addWinner(self, name, amount):
        self.winners[name] = amount

    def setRake(self, amount):
        self.rake = amount

    def endHand(self):
        """to be called after all the bets and winners have been fed in
        calculates how many chips each player is left with and returns net changes"""
        #print "stacks:"
        for name in self.players:
            betsonstreets = [reduce(operator.add, b[name],0) for b in self.bets]
            sumbets = reduce(operator.add, betsonstreets)
            self.endChips[name] = self.startChips[name] - sumbets + self.winners[name]
            self.netWinners[name] = self.endChips[name] - self.startChips[name]
            #print name, self.startChips[name], self.endChips[name]
        return(self.netWinners)
